from insights_analytics_collector import register


@register("config", "1.0", description="CONFIG", config=True)
def config(since, **kwargs):
    return {"version": "1.0"}


@register("json1", "1.1", description="json1")
def json1(**kwargs):
    return {"json1": "True"}


@register("json2", "1.2", description="json2")
def json2(**kwargs):
    return {"json2": "True"}


@register("json3", "1.3", description="json3")
def json3(**kwargs):
    return {"json3": "True"}



import os

from django.utils.timezone import now, timedelta
from insights_analytics_collector import CsvFileSplitter

TIMESTAMP_CSV_LINE_LENGTH = 40


def trivial_slicing(key, last_gather, since, until, **kwargs):
    return [(since, until)]


def one_day_slicing(key, last_gather, since, until, **kwargs):
    since = since.replace(hour=0, minute=0, second=0, microsecond=0)
    until = until.replace(hour=0, minute=0, second=0, microsecond=0)
    start, end = since, None
    while start < until:
        end = min(start + timedelta(days=1), until)
        yield (start, end)
        start = end


def full_sync_slicing(key, last_gather, full_sync_enabled=False, since=None, **kwargs):
    """
    If full_sync_enabled is:
        - True: Yields 10 time slices in 1-day intervals
        - False: Yields slices since 'since' in 1-day intervals
    """
    current_time = now().replace(hour=0, minute=0, second=0, microsecond=0)
    if full_sync_enabled:
        start = current_time - timedelta(days=10)
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        start = since.replace(hour=0, minute=0, second=0, microsecond=0)

    while start < current_time:
        end = start + timedelta(days=1)
        yield (start, end)
        start = end


def csv_generator(full_path, file_name, files_cnt, max_data_size, header, line):
    file_path = get_file_path(full_path, file_name)
    file = CsvFileSplitter(filespec=file_path, max_file_size=max_data_size)

    # create required number of files (decrease by headers - it's CSV)
    file.write(header)
    for _ in range(files_cnt * int(max_data_size / len(line)) - files_cnt):
        file.write(line)

    return file.file_list()


def simple_csv(full_path, file_name, files_cnt, max_data_size):
    """CSVs with line length 10 bytes"""
    header = "Col1,Col2\n"  # 10 chars
    line = "1234,6789\n"  # 10 chars
    return csv_generator(full_path, file_name, files_cnt, max_data_size, header, line)


def timestamp_csv(full_path, file_name, files_cnt, max_data_size, since, until):
    """CSVs with line length 40 bytes"""
    header = "since______________,until______________\n"  # 40 chars
    line = [
        since.strftime("%Y,%m,%d,%H,00,00"),  # 19 chars
        until.strftime("%Y,%m,%d,%H,00,00"),
    ]  # 19 chars
    line = f"{','.join(line)}\n"  # +2 = 40 chars

    return csv_generator(full_path, file_name, files_cnt, max_data_size, header, line)


def get_file_path(path, table):
    return os.path.join(path, table + "_table.csv")


def decode_csv_line(line):
    return line.decode("utf-8").replace("\r", "").replace("\n", "").split(",")


def assert_common_files(files):
    assert "./config.json" in files.keys()
    assert "./manifest.json" in files.keys()
    assert "./data_collection_status.csv" in files.keys()



from insights_analytics_collector import register
from tests.functional.helpers import simple_csv, trivial_slicing


@register("config", "1.0", description="CONFIG", config=True)
def config(since, **kwargs):
    return {"version": "1.0"}


@register(
    "big_table", "1.0", format="csv", description="Testing CSV data - file splitting"
)
def big_table(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "big_table", 10, max_data_size)


@register(
    "big_table_2",
    "1.0",
    format="csv",
    description="Testing CSV data - file splitting 2",
)
def big_table_2(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "big_table", 3, 800)


@register("csv_collection_1", "1.0", format="csv", description="CSV 1")
def csv_collection_1(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "csv_collection_1", 1, max_data_size=100)


@register("csv_collection_2", "1.0", format="csv", description="CSV 2")
def csv_collection_2(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "csv_collection_2", 1, max_data_size=200)


@register("csv_collection_3", "1.0", format="csv", description="CSV 3")
def csv_collection_3(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "csv_collection_3", 1, max_data_size=300)


@register(
    "csv_slicing_1",
    "1.0",
    format="csv",
    description="CSV with slicing 1",
    fnc_slicing=trivial_slicing,
)
def csv_slicing_1(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "csv_slicing_1", 1, max_data_size=100)


@register(
    "csv_slicing_2",
    "1.0",
    format="csv",
    description="CSV with slicing 2",
    fnc_slicing=trivial_slicing,
)
def csv_slicing_2(full_path, max_data_size, **kwargs):
    return simple_csv(full_path, "csv_slicing_2", 2, max_data_size=100)


@register("json_collection_1", "1.0", format="json", description="JSON 1")
def json_collection_1(**kwargs):
    return {"json1": "True"}


@register("json_collection_2", "2.0", format="json", description="JSON 2")
def json_collection_2(**kwargs):
    return {"json2": "True"}



import copy
import os

from django.utils.timezone import now

from .collection import Collection


class CollectionCSV(Collection):
    """Collection for CSV-outputting collecting functions (decorated by @register)
    Collecting functions can return 0+ paths to tmpfiles
    - result of gather() is stored in self.data_filepath
    - In case of multiple files, object clones itself to sub-collections,
      one for each file
    """

    def __init__(self, collector, fnc_collecting):
        super().__init__(collector, fnc_collecting)
        # Large db tables handled by fnc_collecting can be split to multiple files,
        # this is covered by sub_collections
        self.sub_collections = []
        self.data_filepath = None

    def add_to_tar(self, tar):
        """Adds CSV file to the tar(tgz) archive"""
        self.logger.debug(
            f"CollectionCSV._add_to_tar: | {self.key}.csv | Size: {self.data_size()}"
        )
        tar.add(self.target(), arcname=f"./{self.filename}")

    def cleanup(self):
        """Removes CSV files from /tmp"""
        if self.data_filepath and os.path.exists(self.data_filepath):
            os.remove(self.data_filepath)

        for collection in self.sub_collections:
            collection.cleanup()

    def data_size(self):
        """Gets size of tmp csv file. Sub-collections NOT computed."""
        if self.data_filepath is None:
            return 0

        data_size = 0
        try:
            if os.path.exists(self.data_filepath):
                data_size = os.path.getsize(self.data_filepath)
        except OSError as e:
            self.logger.error(f"Can't get size of CSV file: {e}")

        return data_size

    def is_empty(self):
        """
        Leaf checks if data are collected.
        Collection with sub-collections looks for any non-empty sub-collection
        """
        if len(self.sub_collections):
            for collection in self.sub_collections:
                if not collection.is_empty():
                    return False
            return True
        else:
            return self.data_filepath is None

    def target(self):
        """Data attribute specific for this data type"""
        return self.data_filepath

    def update_last_gathered_entries(self, updates_dict):
        """Get last successful gathering or ask sub-collections"""
        if len(self.sub_collections):
            for collection in self.sub_collections:
                collection.update_last_gathered_entries(updates_dict)
        else:
            super().update_last_gathered_entries(updates_dict)

    #
    # Private methods ---------------------------
    #
    def _save_gathering(self, data):
        """
        Saves data (paths to CSV files).
        If there are more files, sub-collections are created
        """
        if isinstance(data, list) and len(data) > 1:
            for fpath in data:
                sub_collection = copy.copy(self)
                sub_collection.sub_collections = []
                sub_collection.data_filepath = fpath
                sub_collection.gathering_successful = True
                self.sub_collections.append(sub_collection)
        elif isinstance(data, list) and len(data) == 1:
            self.data_filepath = data[0]
        elif isinstance(data, str):
            self.data_filepath = data

    def _set_gathering_finished(self):
        _now = now()
        self.gathering_finished_at = _now
        for sub_collection in self.sub_collections:
            sub_collection.gathering_finished_at = _now



from insights_analytics_collector import register
from tests.functional.helpers import (
    TIMESTAMP_CSV_LINE_LENGTH,
    full_sync_slicing,
    one_day_slicing,
    timestamp_csv,
)


@register("config", "1.0", description="CONFIG", config=True)
def config(since, **kwargs):
    return {"version": "1.0"}


@register(
    "csv_one_day_slicing_1",
    "1.0",
    format="csv",
    description="CSVs splitted by date",
    fnc_slicing=one_day_slicing,
)
def csv_one_day_slicing_1(since, full_path, until, **kwargs):
    return timestamp_csv(
        full_path,
        "csv_one_day_slicing_1",
        1,
        2 * TIMESTAMP_CSV_LINE_LENGTH,
        since=since,
        until=until,
    )


@register(
    "csv_one_day_slicing_2",
    "1.0",
    format="csv",
    description="CSVs splitted by size and date",
    fnc_slicing=one_day_slicing,
)
def csv_one_day_slicing_2(since, full_path, until, **kwargs):
    return timestamp_csv(
        full_path,
        "csv_one_day_slicing_2",
        2,
        2 * TIMESTAMP_CSV_LINE_LENGTH,
        since=since,
        until=until,
    )


@register(
    "csv_full_sync_slicing_1",
    "1.0",
    format="csv",
    description="CSVs splitted by date",
    fnc_slicing=full_sync_slicing,
    full_sync_interval_days=5,
)
def csv_full_sync_slicing_1(since, full_path, until, **kwargs):
    return timestamp_csv(
        full_path,
        "csv_full_sync_slicing_1",
        1,
        2 * TIMESTAMP_CSV_LINE_LENGTH,
        since=since,
        until=until,
    )



from abc import abstractmethod

from django.utils.timezone import now, timedelta


class Collection:
    """Wrapper for gathering function from Collector.collector_module
    Functions decorated with @register are wrapped by kind of this object.
    """

    COLLECTION_TYPE_CONFIG = "config"
    COLLECTION_TYPE_JSON = "json"
    COLLECTION_TYPE_CSV = "csv"

    def __init__(self, collector, fnc_collecting):
        self.collector = collector
        self.logger = collector.logger
        self.fnc_collecting = fnc_collecting
        self.fnc_slicing = fnc_collecting.__insights_analytics_fnc_slicing__
        self.is_config = fnc_collecting.__insights_analytics_config__

        self.description = fnc_collecting.__insights_analytics_description__ or ""
        self.key = fnc_collecting.__insights_analytics_key__
        self.shipping_group = fnc_collecting.__insights_analytics_shipping_group__
        self.version = fnc_collecting.__insights_analytics_version__

        self.data_type = fnc_collecting.__insights_analytics_type__
        self.filename = f"{self.key}.{self.data_type}"
        # either since/until or full sync(if enabled)
        self.since = None  # set by Collector._create_collections()
        self.until = None  # set by Collector._create_collections()
        self.full_sync_enabled = self._is_full_sync_enabled(
            fnc_collecting.__insights_analytics_full_sync_interval_days__
        )

        self.gathering_started_at = None
        self.gathering_finished_at = None
        self.gathering_successful = None
        self.last_gathered_entry = self.collector.last_gathered_entry_for(self.key)

    @abstractmethod
    def add_to_tar(self, tar):
        pass

    def cleanup(self):
        """There is an action only for CollectionCSV"""
        pass

    @abstractmethod
    def data_size(self):
        pass

    def gather(self, max_data_size):
        self.gathering_started_at = now()

        try:
            # More collections with the same key (and different since/until)
            # have the same file names => overwriting! [error]
            result = self.fnc_collecting(
                since=self.since,
                until=self.until,
                max_data_size=max_data_size,
                full_path=self.collector.gather_dir,
                collection_type=self.collector.collection_type,
            )
            self._save_gathering(result)

            self.gathering_successful = True
        except Exception as e:
            self.logger.exception(f"Could not generate metric {self.filename}: {e}")
            self.gathering_successful = False
        finally:
            self._set_gathering_finished()

    @abstractmethod
    def is_empty(self):
        pass

    def slices(self):
        since = self.collector.gather_since
        until = self.collector.gather_until
        last_gather = self.collector.last_gather
        # These slicer functions may return a generator. The `since` parameter is
        # allowed to be None, and will fall back to LAST_ENTRIES[key] or to
        # LAST_GATHER (truncated appropriately to match the 4-week limit).
        #
        # Or it can force full table sync if interval is given
        if self.fnc_slicing:
            if self.full_sync_enabled:
                slices = self.fnc_slicing(self.key, last_gather, full_sync_enabled=True)
            else:
                slices = self.fnc_slicing(
                    self.key, last_gather, since=since, until=until
                )
        else:
            slices = [(self._gather_since(), self._gather_until())]

        return slices

    def ship_immediately(self):
        """
        Collection with fnc_slicing has to be shipped immediately.
        It may gather to the same file(s) as previous slice.
        Keeping more files is not wanted because of their potential size
        """
        return self.fnc_slicing is not None

    @abstractmethod
    def target(self):
        """Data attribute specific for collection"""
        pass

    def update_last_gathered_entries(self, updates_dict):
        if self.key in updates_dict["locked"]:
            return

        if self.gathering_successful:
            self._update_last_gathered_key(updates_dict, self.key, self.until)

            if self.full_sync_enabled:
                self._update_last_gathered_key(
                    updates_dict, f"{self.key}_full", self.gathering_finished_at
                )
        else:
            # collections are ordered by time slices.
            # in case of error all collections with newer timestamp are ignored
            updates_dict["locked"].add(self.key)

    #
    # Private methods ---------------------------
    #
    @staticmethod
    def _update_last_gathered_key(updates_dict, key, timestamp):
        previous = updates_dict["keys"].get(key, None)
        if previous is None:
            updates_dict["keys"][key] = timestamp
        else:
            updates_dict["keys"][key] = max(previous, timestamp)

    def _is_full_sync_enabled(self, interval_days):
        if not interval_days:
            return False

        last_full_sync = self.collector.last_gathered_entry_for(f"{self.key}_full")
        return not last_full_sync or last_full_sync < now() - timedelta(
            days=interval_days
        )

    def _gather_since(self):
        """Start of gathering based on settings excluding slices"""
        last_entry = max(
            self.last_gathered_entry or self.collector.last_gather,
            self.collector.gather_until - timedelta(weeks=4),
        )
        return self.collector.gather_since or last_entry

    def _gather_until(self):
        """End of gathering based on settings excluding slices"""
        return self.collector.gather_until

    @abstractmethod
    def _save_gathering(self, data):
        pass

    def _set_gathering_finished(self):
        self.gathering_finished_at = now()



from insights_analytics_collector import register
from tests.functional.helpers import simple_csv, trivial_slicing


@register("config", "1.0", description="CONFIG", config=True)
def config(since, **kwargs):
    return {"version": "1.0"}


@register("simple_json1", "1.0", description="json1")
def simple_json1(**kwargs):
    return {"simple_json": "True"}


@register("csv_no_slicing_1-2x", "1.0", format="csv", description="CSV No slicing 1")
def csv_no_slicing1(full_path, **kwargs):
    return simple_csv(full_path, "csv_no_slicing_1-2x", 2, 100)


@register(
    "csv_with_slicing_1-5x",
    "1.0",
    format="csv",
    description="CSV With Slicing 1a",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing1a(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_1-5x", 2, 100)


@register(
    "csv_with_slicing_1-5x",
    "1.0",
    format="csv",
    description="CSV With Slicing 1b",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing1b(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_1-5x", 3, 100)


@register("csv_no_slicing_2-1x", "1.0", format="csv", description="CSV No slicing 2")
def csv_no_slicing2(full_path, **kwargs):
    return simple_csv(full_path, "csv_no_slicing_2-1x", 1, 100)


@register(
    "csv_with_slicing_2-3x",
    "1.0",
    format="csv",
    description="CSV With Slicing 2a",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing2a(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_2-3x", 2, 100)


@register("csv_no_slicing_3-10x", "1.0", format="csv", description="CSV No slicing 3")
def csv_no_slicing3(full_path, **kwargs):
    return simple_csv(full_path, "csv_no_slicing_3-10x", 10, 100)


@register(
    "csv_with_slicing_3-2x",
    "2.0",
    format="csv",
    description="CSV With Slicing 3",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing3(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_3-2x", 2, 100)


@register("csv_no_slicing_4-12x", "1.0", format="csv", description="CSV No slicing 3")
def csv_no_slicing4(full_path, **kwargs):
    return simple_csv(full_path, "csv_no_slicing_4-12x", 12, 100)


@register(
    "csv_with_slicing_2-3x",
    "2.0",
    format="csv",
    description="CSV With Slicing 2b",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing2b(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_2-3x", 1, 100)


@register(
    "csv_with_slicing_4-3x",
    "2.0",
    format="csv",
    description="CSV With Slicing 4",
    fnc_slicing=trivial_slicing,
)
def csv_with_slicing4(full_path, **kwargs):
    return simple_csv(full_path, "csv_with_slicing_4-3x", 3, 100)



from .collection_csv import CollectionCSV
from .collection_json import CollectionJSON
from .collector import Collector
from .csv_file_splitter import CsvFileSplitter
from .decorators import register, slicing
from .package import Package

__all__ = [
    "Collector",
    "Package",
    "CsvFileSplitter",
    "CollectionCSV",
    "CollectionJSON",
    "register",
    "slicing",
]



import base64
import json
import os
import pathlib
import tarfile
from abc import abstractmethod

import requests


class Package:
    """
    Abstract class
    Package serves for managing one tarball and shipping it to the cloud.
    Abstract methods has to be implemented, as well as:
    - CERT_PATH - path to auth certificate (for POST request to cloud), if not development mode
    - PAYLOAD_CONTENT_TYPE - registered in ingress-service in cloud
    - MAX_DATA_SIZE - defaults to 200MB (upload limit is 100MB, so it expects 50% compression rate)

    See the README.md and tests/functional/test_gathering.py to see how are packages used
    """

    CERT_PATH = "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem"
    # i.e. "application/vnd.redhat.tower.tower_payload+tgz"
    PAYLOAD_CONTENT_TYPE = "application/vnd.redhat.TODO+tgz"

    SHIPPING_AUTH_USERPASS = "user-pass"
    SHIPPING_AUTH_S3_USERPASS = "user-pass-s3"
    SHIPPING_AUTH_IDENTITY = "x-rh-identity"  # Development mode only
    SHIPPING_AUTH_CERTIFICATES = "mutual-tls"  # Mutual TLS

    DEFAULT_RHSM_CERT_FILE = "/etc/pki/consumer/cert.pem"
    DEFAULT_RHSM_KEY_FILE = "/etc/pki/consumer/key.pem"

    """
    Some tables can be *very* large, and we have a 100MB upload limit.

    Split large table dumps at dump time into a series of files.
    """
    MAX_DATA_SIZE = 200 * 1048576

    def __init__(self, collector):
        self.collector = collector
        self.collections = []
        self.collection_keys = []
        self.data_collection_status = self.collector.collection_data_status_class()(
            self.collector, self
        )
        self.logger = collector.logger
        self.manifest = collector.collection_manifest_class()(collector)
        self.processed = False
        self.shipping_successful = None
        self.tar_path = None
        self.total_data_size = 0

    @classmethod
    def max_data_size(cls):
        return cls.MAX_DATA_SIZE

    def add_collection(self, collection):
        self.collections.append(collection)
        self.collection_keys.append(collection.key)
        self.total_data_size = self.total_data_size + collection.data_size()

    def is_key_used(self, key):
        return key in self.collection_keys

    def delete_collected_files(self):
        for collection in self.collections:
            collection.cleanup()

    @abstractmethod
    def get_ingress_url(self):
        """URL of cloud's upload URL"""
        pass

    @abstractmethod
    def get_s3_configured(self):
        """URL of cloud's upload URL"""
        pass

    def has_free_space(self, requested_size):
        return self.total_data_size + requested_size <= self.max_data_size()

    def is_shipping_configured(self):
        if not self.tar_path:
            self.logger.error("Insights for Ansible Automation Platform TAR not found")
            return False

        if not os.path.exists(self.tar_path):
            self.logger.error(
                f"Insights for Ansible Automation Platform TAR {self.tar_path} not found"
            )
            return False

        if "Error:" in str(self.tar_path):
            return False

        if self.shipping_auth_mode() == self.SHIPPING_AUTH_USERPASS:
            if not self.get_ingress_url():
                self.logger.error("AUTOMATION_ANALYTICS_URL is not set")
                return False

            if not self._get_rh_user():
                self.logger.error("REDHAT_USERNAME is not set")
                return False

            if not self._get_rh_password():
                self.logger.error("REDHAT_PASSWORD is not set")
                return False

        if self.shipping_auth_mode() == self.SHIPPING_AUTH_S3_USERPASS:
            if not self.get_s3_configured():
                self.logger.error("S3 configuration is not set")
                return False

            if not self._get_rh_user():
                self.logger.error("aws_access_key_id is not set")
                return False

            if not self._get_rh_password():
                self.logger.error("aws_secret_access_key is not set")
                return False

            if not self._get_rh_region():
                self.logger.error("aws_region is not set")
                return False

            if not self._get_rh_bucket():
                self.logger.error("aws_bucket is not set")
                return False

        return True

    def make_tgz(self):
        target = self.collector.tmp_dir.parent
        try:
            tarname_base = self._tarname_base()
            path = pathlib.Path(target)
            index = len(list(path.glob(f"{tarname_base}-*.*")))
            tarname = f"{tarname_base}-{index}.tar.gz"

            with tarfile.open(target.joinpath(tarname), "w:gz") as f:
                for collection in self.collections:
                    self._collection_to_tar(f, collection)

                self._config_to_tar(f)

                self._data_collection_status_to_tar(f)

                self._manifest_to_tar(f)

                self.tar_path = f.name
            return True
        except Exception as e:
            self.logger.exception(f"Failed to write analytics archive file: {e}")
            return False

    def ship(self):
        """
        Ship gathered metrics to the Insights API
        """
        if not self.is_shipping_configured():
            self.shipping_successful = False
            return False

        self.logger.debug(f"shipping analytics file: {self.tar_path}")

        with open(self.tar_path, "rb") as f:
            files = {
                "file": (
                    os.path.basename(self.tar_path),
                    f,
                    self._payload_content_type(),
                )
            }
            s = requests.Session()
            if self.shipping_auth_mode() == self.SHIPPING_AUTH_CERTIFICATES:
                # as a single file (containing the private key and the certificate) or
                # as a tuple of both files paths (cert_file, keyfile)
                s.cert = self._get_client_certificates()

            s.headers = self._get_http_request_headers()
            s.headers.pop("Content-Type")

            if self.shipping_auth_mode() == self.SHIPPING_AUTH_IDENTITY:
                s.headers["x-rh-identity"] = self._get_x_rh_identity()

            url = self.get_ingress_url()
            self.shipping_successful = self._send_data(url, files, s)

        return self.shipping_successful

    def shipping_auth_mode(self):
        return self.SHIPPING_AUTH_USERPASS

    def update_last_gathered_entries(self, updates_dict):
        if self.shipping_successful:
            for collection in self.collections:
                collection.update_last_gathered_entries(updates_dict)

    #
    # Private methods ---------------------------
    #

    def _collection_to_tar(self, tar, collection):
        try:
            if not collection.is_empty():
                collection.add_to_tar(tar)
                self.manifest.add_collection(collection)
        except Exception as e:
            self.logger.exception(
                f"Could not generate metric {collection.filename}: {e}"
            )
            return None

    def _send_data(self, url, files, session):
        if self.shipping_auth_mode() == self.SHIPPING_AUTH_USERPASS:
            response = session.post(
                url,
                files=files,
                verify=self.CERT_PATH,
                auth=(self._get_rh_user(), self._get_rh_password()),
                headers=session.headers,
                timeout=(31, 31),
            )
        else:
            response = session.post(
                url, files=files, headers=session.headers, timeout=(31, 31)
            )

        # Accept 2XX status_codes
        if response.status_code >= 300:
            self.logger.error(
                "Upload failed with status {}, {}".format(
                    response.status_code, response.text
                )
            )
            return False

        return True

    def _config_to_tar(self, tar):
        if self.collector.collections["config"] is None:
            self.logger.error(
                "'config' collector data is missing, and is required to ship."
            )
            return False
        else:
            self._collection_to_tar(tar, self.collector.collections["config"])

        return True

    @abstractmethod
    def _get_http_request_headers(self):
        """Optional HTTP headers for POST request to get_ingress_url() URL
        :return: dict()
        """
        pass

    @abstractmethod
    def _get_rh_user(self):
        """Auth: username for HTTP POST request to cloud.
        shipping_auth_mode() must return SHIPPING_AUTH_USERPASS (default)
        """
        pass

    @abstractmethod
    def _get_rh_password(self):
        """Auth: password for HTTP POST request to cloud.
        shipping_auth_mode() must return SHIPPING_AUTH_USERPASS (default)
        """
        pass

    @abstractmethod
    def _get_rh_region(self):
        """s3: The region that boto3 will connect to"""
        pass

    @abstractmethod
    def _get_rh_bucket(self):
        """s3: The bucket that boto3 will use"""
        pass

    def _get_client_certificates(self):
        """Auth: get client certificate and key, by default we use the RHSM certs
        :return: string or tuple of 2 strings
        """
        return (self.DEFAULT_RHSM_CERT_FILE, self.DEFAULT_RHSM_KEY_FILE)

    def _get_x_rh_identity(self):
        """Auth: x-rh-identity header for HTTP POST request to cloud
        Optional, if shipping_auth_mode() redefined to SHIPPING_AUTH_IDENTITY
        """
        identity = {
            "identity": {
                "type": "User",
                "account_number": "0000001",
                "user": {"is_org_admin": True},
                "internal": {"org_id": "000001"},
            }
        }
        identity = base64.b64encode(json.dumps(identity).encode("utf8"))
        return identity

    def _data_collection_status_to_tar(self, tar):
        try:
            self.data_collection_status.gather(None)
            self.data_collection_status.add_to_tar(tar)
            self.manifest.add_collection(self.data_collection_status)
        except Exception as e:
            self.logger.exception(
                f"Could not generate {self.data_collection_status.filename}: {e}"
            )

    def _manifest_to_tar(self, tar):
        try:
            self.manifest.gather(None)
            self.manifest.add_to_tar(tar)
            self.add_collection(self.manifest)
        except Exception as e:
            self.logger.exception(f"Could not generate {self.manifest.filename}: {e}")

    def _payload_content_type(self):
        return self.PAYLOAD_CONTENT_TYPE

    def _tarname_base(self):
        timestamp = self.collector.gather_until
        return f'analytics-{timestamp.strftime("%Y-%m-%d-%H%M%S%z")}'



def register(
    key,
    version,
    description=None,
    format="json",
    config=False,
    fnc_slicing=None,
    shipping_group="default",
    full_sync_interval_days=None,
):
    """
    A decorator used to register a function as a metric collector.

    Decorated functions should do the following based on format:
    - json: return JSON-serializable objects.
    - csv: write CSV data to a filename named 'key'

    :param output_type - 'data' or 'file_paths'

    @register('projects_by_scm_type', 1)
    def projects_by_scm_type():
        return {'git': 5, 'svn': 1}
    """

    def decorate(f):
        f.__insights_analytics_key__ = key
        f.__insights_analytics_version__ = version
        f.__insights_analytics_description__ = description
        f.__insights_analytics_type__ = format  # CSV/JSON
        f.__insights_analytics_config__ = config  # config
        f.__insights_analytics_fnc_slicing__ = fnc_slicing
        f.__insights_analytics_shipping_group__ = shipping_group
        f.__insights_analytics_full_sync_interval_days__ = full_sync_interval_days

        return f

    return decorate


def slicing(default=False):
    def decorate(f):
        f.__insights_analytics_slicing_name__ = f.__name__
        f.__insights_analytics_slicing_default__ = default
        return f

    return decorate



import json
import logging
import tarfile

import pytest
import tests.functional.collector_module
import tests.functional.collector_module2
import tests.functional.collector_module3
from tests.classes.analytics_collector import AnalyticsCollector
from tests.functional.helpers import assert_common_files, decode_csv_line


@pytest.fixture
def collector(mocker):
    collector = AnalyticsCollector(
        collector_module=tests.functional.collector_module,
        collection_type=AnalyticsCollector.DRY_RUN,
    )
    mocker.patch.object(collector, "_is_valid_license", return_value=True)

    return collector


def test_missing_config(mocker, collector):
    mock_logger = mocker.patch.object(collector, "logger")

    tgz_files = collector.gather(subset=["json_collection_1", "json_collection_2"])

    assert tgz_files is None
    mock_logger.log.assert_called_with(
        logging.ERROR, "'config' collector data is missing"
    )


def test_json_collections(collector):
    tgz_files = collector.gather(
        subset=["config", "json_collection_1", "json_collection_2"]
    )

    assert len(tgz_files) == 1

    files = {}
    with tarfile.open(tgz_files[0], "r:gz") as archive:
        for member in archive.getmembers():
            files[member.name] = archive.extractfile(member)

        assert_common_files(files)
        assert "./json_collection_1.json" in files.keys()
        assert "./json_collection_2.json" in files.keys()

        assert json.loads(files["./config.json"].read()) == {"version": "1.0"}
        assert json.loads(files["./json_collection_1.json"].read()) == {"json1": "True"}
        assert json.loads(files["./json_collection_2.json"].read()) == {"json2": "True"}

    collector._gather_cleanup()


def test_small_csvs(collector):
    tgz_files = collector.gather(
        subset=["config", "csv_collection_1", "csv_collection_2", "csv_collection_3"]
    )

    assert len(tgz_files) == 1

    files = {}
    with tarfile.open(tgz_files[0], "r:gz") as archive:
        for member in archive.getmembers():
            files[member.name] = archive.extractfile(member)

        assert_common_files(files)
        assert "./csv_collection_1.csv" in files.keys()
        assert "./csv_collection_2.csv" in files.keys()
        assert "./csv_collection_3.csv" in files.keys()

        # length defined by @registered function
        assert len(files["./csv_collection_1.csv"].read()) == 100
        assert len(files["./csv_collection_2.csv"].read()) == 200
        assert len(files["./csv_collection_3.csv"].read()) == 300

    collector._gather_cleanup()


def test_jsons_with_csvs_with_slicing(collector):
    tgz_files = collector.gather(
        subset=[
            "config",
            "json_collection_1",
            "json_collection_2",
            "csv_slicing_1",
            "csv_slicing_2",
        ]
    )

    assert len(tgz_files) == 3

    for i in range(len(tgz_files)):
        files = {}
        with tarfile.open(tgz_files[i], "r:gz") as archive:
            for member in archive.getmembers():
                files[member.name] = archive.extractfile(member)

            assert_common_files(files)
            if i == 0:
                assert "./json_collection_1.json" in files.keys()
                assert "./json_collection_2.json" in files.keys()
                assert "./csv_slicing_1.csv" in files.keys()
            if i == 1:
                assert "./csv_slicing_2.csv" in files.keys()
            if i == 2:
                assert "./csv_slicing_2.csv" in files.keys()


def test_one_csv_collection_splitted_by_size(collector):
    tgz_files = collector.gather(subset=["config", "big_table"])

    assert len(tgz_files) == 10

    for i in range(len(tgz_files)):
        files = {}
        with tarfile.open(tgz_files[i], "r:gz") as archive:
            for member in archive.getmembers():
                files[member.name] = archive.extractfile(member)

            assert_common_files(files)
            assert len(files.keys()) == 1 + _common_files_count()
            assert "./big_table.csv" in files.keys()
            assert len(files["./big_table.csv"].read()) == 1000

    collector._gather_cleanup()


def test_multiple_collections_multiple_tarballs(mocker, collector):
    mocker.patch("tests.classes.package.Package.MAX_DATA_SIZE", 1000)

    tgz_files = collector.gather(
        subset=["config", "big_table_2", "csv_collection_1", "csv_collection_2"]
    )

    assert len(tgz_files) == 3

    for i in range(len(tgz_files)):
        files = {}
        with tarfile.open(tgz_files[i], "r:gz") as archive:
            for member in archive.getmembers():
                files[member.name] = archive.extractfile(member)

            assert_common_files(files)
            if i == 0:
                assert len(files.keys()) == 2 + _common_files_count()
                assert "./big_table_2.csv" in files.keys()
                assert "./csv_collection_1.csv" in files.keys()
            elif i == 1:
                assert len(files.keys()) == 2 + _common_files_count()
                assert "./big_table_2.csv" in files.keys()
                assert "./csv_collection_2.csv" in files.keys()
            elif i == 2:
                assert len(files.keys()) == 1 + _common_files_count()
                assert "./big_table_2.csv" in files.keys()

    collector._gather_cleanup()


def test_multiple_collections_and_distributions(collector):
    """
    - JSONs start at index 0
    - CSVs with no slicing start at index 0
    - CSVs with slicing start after index next to previous slice
    """
    collector.collector_module = tests.functional.collector_module3
    tgz_files = collector.gather()

    assert len(tgz_files) == 13

    for i in range(len(tgz_files)):
        files = {}
        with tarfile.open(tgz_files[i], "r:gz") as archive:
            for member in archive.getmembers():
                files[member.name] = archive.extractfile(member)

            assert_common_files(files)
            if i == 0:
                assert "./simple_json1.json" in files.keys()
            else:
                assert "./simple_json1.json" not in files.keys()

            # CSVs with no slicing start at index 0
            if 0 <= i <= 1:
                assert "./csv_no_slicing_1-2x.csv" in files.keys()
            else:
                assert "./csv_no_slicing_1-2x.csv" not in files.keys()

            if i == 0:
                assert "./csv_no_slicing_2-1x.csv" in files.keys()
            else:
                assert "./csv_no_slicing_2-1x.csv" not in files.keys()

            if 0 <= i <= 9:
                assert "./csv_no_slicing_3-10x.csv" in files.keys()
            else:
                assert "./csv_no_slicing_3-10x.csv" not in files.keys()

            if 0 <= i <= 11:
                assert "./csv_no_slicing_4-12x.csv" in files.keys()
            else:
                assert "./csv_no_slicing_4-12x.csv" not in files.keys()

            # CSVs with slicing start after index next to previous slice
            if 0 <= i <= 4:
                assert "./csv_with_slicing_1-5x.csv" in files.keys()
            else:
                assert "./csv_with_slicing_1-5x.csv" not in files.keys()

            if 5 <= i <= 7:
                assert "./csv_with_slicing_2-3x.csv" in files.keys()
            else:
                assert "./csv_with_slicing_2-3x.csv" not in files.keys()

            if 8 <= i <= 9:
                assert "./csv_with_slicing_3-2x.csv" in files.keys()
            else:
                assert "./csv_with_slicing_3-2x.csv" not in files.keys()

            if 10 <= i <= 12:
                assert "./csv_with_slicing_4-3x.csv" in files.keys()
            else:
                assert "./csv_with_slicing_4-3x.csv" not in files.keys()


def test_manifest_and_status(collector):
    collector.collector_module = tests.functional.collector_module2
    tgz_files = collector.gather()

    assert len(tgz_files) == 1

    files = {}
    with tarfile.open(tgz_files[0], "r:gz") as archive:
        for member in archive.getmembers():
            files[member.name] = archive.extractfile(member)

        assert_common_files(files)
        assert len(files.keys()) == 3 + _common_files_count()

        assert json.loads(files["./manifest.json"].read()) == {
            "config.json": "1.0",
            "data_collection_status.csv": "1.0",
            "json1.json": "1.1",
            "json2.json": "1.2",
            "json3.json": "1.3",
        }

        _assert_data_collection_status(files["./data_collection_status.csv"])

    collector._gather_cleanup()


def _assert_data_collection_status(status_file):
    lines = status_file.readlines()
    files = ["json1.json", "json2.json", "json3.json"]
    assert len(lines) == len(files) + 1  # +1 == header

    fieldnames = [
        "collection_start_timestamp",
        "since",
        "until",
        "file_name",
        "status",
        "elapsed",
    ]
    header = lines.pop(0)
    assert decode_csv_line(header) == fieldnames
    for line in lines:
        row = decode_csv_line(line)
        assert len(row) == len(fieldnames)
        assert row[4] == "ok"  # status
        assert row[5] == "0"  # elapsed
        files.pop(files.index(row[3]))

    assert len(files) == 0


def _common_files_count():
    return 3



import csv
import os

from .collection_csv import CollectionCSV
from .decorators import register


class CollectionDataStatus(CollectionCSV):
    def __init__(self, collector, package):
        super().__init__(collector, self.data_collection_status)

        self.package = package

    @register(
        "data_collection_status",
        "1.0",
        format="csv",
        description="Data collection status",
    )
    def data_collection_status(self, full_path, **kwargs):
        file_path = os.path.join(full_path, self.filename)
        with open(file_path, "w", newline="") as csvfile:
            fieldnames = [
                "collection_start_timestamp",
                "since",
                "until",
                "file_name",
                "status",
                "elapsed",
            ]
            writer = csv.DictWriter(csvfile, delimiter=",", fieldnames=fieldnames)
            writer.writeheader()

            for collection in self.package.collections:
                status = "ok" if collection.gathering_successful else "failed"
                elapsed = 0
                if collection.gathering_started_at and collection.gathering_finished_at:
                    elapsed = (
                        collection.gathering_finished_at
                        - collection.gathering_started_at
                    ).seconds

                writer.writerow(
                    {
                        "collection_start_timestamp": collection.gathering_started_at,
                        "since": collection.since,
                        "until": collection.until,
                        "file_name": collection.filename,
                        "status": status,
                        "elapsed": elapsed,
                    }
                )
        return [file_path]



import datetime
import tarfile

import pytest
import pytz
import tests.functional.collector_module4_slicing
from django.utils.timezone import now, timedelta
from tests.classes.analytics_collector import AnalyticsCollector
from tests.functional.helpers import assert_common_files, decode_csv_line


@pytest.fixture
def collector(mocker):
    """
    Collector with registered module `collector_module4_slicing`.
    It's designed to use non-trivial slicing functions
    """
    collector = AnalyticsCollector(
        collector_module=tests.functional.collector_module4_slicing,
        collection_type=AnalyticsCollector.DRY_RUN,
    )
    mocker.patch.object(collector, "_is_valid_license", return_value=True)

    return collector


def test_slices_by_date(collector):
    """
    Slicing by fnc_slicing function `one_day_slicing`.
    Splits data to 10 days/files.
    """
    days_to_collect = 10

    until = now().replace(hour=0, minute=0, second=0, microsecond=0)
    since = until - timedelta(days=days_to_collect)

    tgz_files = collector.gather(
        subset=["config", "csv_one_day_slicing_1"], since=since, until=until
    )

    assert len(tgz_files) == days_to_collect

    idx = 0
    while since < until:
        files = {}
        with tarfile.open(tgz_files[idx], "r:gz") as archive:
            for member in archive.getmembers():
                files[member.name] = archive.extractfile(member)

            assert_common_files(files)
            assert "./csv_one_day_slicing_1.csv" in files.keys()

            lines = files["./csv_one_day_slicing_1.csv"].readlines()
            _header = lines.pop(0)
            row = decode_csv_line(lines[0])

            csv_since = datetime.datetime(
                int(row[0]),
                int(row[1]),
                int(row[2]),
                int(row[3]),
                int(row[4]),
                int(row[5]),
                tzinfo=pytz.utc,
            )
            csv_until = datetime.datetime(
                int(row[6]),
                int(row[7]),
                int(row[8]),
                int(row[9]),
                int(row[10]),
                int(row[11]),
                tzinfo=pytz.utc,
            )

            assert csv_since == since
            assert csv_until == since + timedelta(days=1)

        idx += 1
        since += timedelta(days=1)


def test_slices_by_date_and_size(collector):
    """
    Slicing by fnc_slicing function `one_day_slicing()`.
    Splits data to 10 days/files.
    Also splits each day to 2 files by size (CsvFileSplitter)
    => total 20 files
    """
    days_to_collect = 10

    until = now().replace(hour=0, minute=0, second=0, microsecond=0)
    since = until - timedelta(days=days_to_collect)

    tgz_files = collector.gather(
        subset=["config", "csv_one_day_slicing_2"], since=since, until=until
    )

    assert len(tgz_files) == days_to_collect * 2


@pytest.mark.parametrize("last_sync_days_ago", [4, 6])
def test_slices_by_full_sync(mocker, collector, last_sync_days_ago):
    """
    In the collector method `csv_full_sync_slicing_1()` there is 5 days interval for full sync
    if `last_sync_days_ago` is:
    - 4 days ago, it uses `since` (7 days ago)
    - 6 days ago, it uses slicing interval (10 days ago) (in `full_sync_slicing()`)

    """
    last_gathered_entries = {
        "csv_full_sync_slicing_1": (now() - timedelta(days=7)).replace(
            hour=0, minute=0, second=0, microsecond=0
        ),
        "csv_full_sync_slicing_1_full": now() - timedelta(days=last_sync_days_ago),
    }
    mocker.patch.object(
        collector, "_load_last_gathered_entries", return_value=last_gathered_entries
    )

    tgz_files = collector.gather(
        subset=["config", "csv_full_sync_slicing_1"],
        since=last_gathered_entries["csv_full_sync_slicing_1"],
    )

    if last_sync_days_ago == 4:
        assert len(tgz_files) == 7
    else:
        assert len(tgz_files) == 10



import io
import os

from .package import Package


class CsvFileSplitter(io.StringIO):
    """Helper for writing big data into multiple files splitted by size.
    Expects data written in CSV format (first line is header)
    Could be called from function decorated by @register (see Collector).
    :param max_file_size: determined by decorated function's attribute "max_data_size"
    """

    def __init__(
        self, filespec=None, max_file_size=Package.MAX_DATA_SIZE, *args, **kwargs
    ):
        self.max_file_size = max_file_size
        self.filespec = filespec
        self.files = []
        self.currentfile = None
        self.header = None
        self.counter = 0
        self.cycle_file()

    def cycle_file(self):
        """Closes current file, opens new one and writes CSV header"""
        if self.currentfile:
            self.currentfile.close()
        self.counter = 0
        fname = "{}_split{}".format(self.filespec, len(self.files))
        self.currentfile = open(fname, "w", encoding="utf-8")
        self.files.append(fname)
        if self.header:
            self.counter += self.currentfile.write("{}\n".format(self.header))

    def file_list(self):
        """Returns list of written files"""
        self.currentfile.close()
        # Check for an empty dump
        if len(self.header) + 1 == self.counter:
            os.remove(self.files[-1])
            self.files = self.files[:-1]
        # If we only have one file, remove the suffix
        if len(self.files) == 1:
            filename = self.files.pop()
            new_filename = filename.replace("_split0", "")
            os.rename(filename, new_filename)
            self.files.append(new_filename)
        return self.files

    def write(self, s):
        """Writes to file and creates new one if file exceedes threshold"""
        if not self.header:
            self.header = s[: s.index("\n")]
        self.counter += self.currentfile.write(s)
        if self.counter >= self.max_file_size:
            self.cycle_file()



import io
import json
import tarfile

from .collection import Collection


class CollectionJSON(Collection):
    """Collection for JSON-outputting collecting functions (decorated by @register)
    Collecting functions returns dict() convertable to JSON
    - result of gather() is stored in self.data
    """

    def __init__(self, collector, func):
        super().__init__(collector, func)
        self.data = None  # gathered data

    def _save_gathering(self, data):
        self.data = json.dumps(data)

    def data_size(self):
        return len(self.data) if self.data else 0

    def is_empty(self):
        return self.data is None

    def target(self):
        """Data attribute specific for this data type"""
        return self.data

    def add_to_tar(self, tar):
        """Adds JSON data to TAR(tgz) archive"""
        buf = self.target().encode("utf-8")
        self.logger.debug(
            f"CollectionJSON._add_to_tar: | {self.key}.json | Size: {self.data_size()}"
        )
        info = tarfile.TarInfo(f"./{self.filename}")
        info.size = len(buf)
        info.mtime = self.collector.gather_until.timestamp()
        tar.addfile(info, fileobj=io.BytesIO(buf))



import os

from setuptools import find_packages, setup

__version__ = "0.3.2"

# We use the README as the long_description
readme_path = os.path.join(os.path.dirname(__file__), "README.md")


setup(
    name="insights-analytics-collector",
    version=__version__,
    url="https://github.com/RedHatInsights/insights-analytics-collector/",
    author="Martin Slemr",
    author_email="mslemr@redhat.com",
    description="Collector Package for Insights for AAP",
    long_description=open(readme_path).read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    license="Apache",
    zip_safe=False,
    packages=find_packages(),
    include_package_data=False,
    install_requires=["django", "requests"],
    tests_require=["pytest", "pytest-mock", "pytz"],
)



import contextlib
import hashlib
import inspect
import logging
import os
import pathlib
import shutil
import tempfile
from abc import abstractmethod

from django.utils.timezone import now, timedelta

from .collection import Collection
from .collection_csv import CollectionCSV
from .collection_data_status import CollectionDataStatus
from .collection_json import CollectionJSON
from .collection_manifest import CollectionManifest
from .package import Package


class Collector:
    """Abstract class. The Collector is an entry-point for gathering data
       from awx to cloud.
    Abstract and following methods has to be implemented:
    - _package_class() - reference to your implementation of Package
    - _collection_json_class() - optional, if class inherited fromCollectionJSON is used
    - _collection_csv_class() - optional, if class inherited from CollectionCSV is used

    There are several params:
    - collection_type:
      - manual/scheduled - data are gathered and shipped, local timestamps about gathering are updated
      - dry-run - data are gathered, but not shipped, tarballs from /tmp not deleted (testing mode)
    - collector_module: module with functions with decorator `@register` - they define what data are collected
      - collector functions are wrapped by kind of Collection object
      - Collections are grouped by Package, and Packages are creating tarballs and shipping them.
    - logger: logging.logger

    Collector is an abstract class, example of implementation is in tests/classes

    Data are gathered maximally 4 weeks ago and can be set to less (see gather(since, until,..))
    """

    MANUAL_COLLECTION = "manual"
    DRY_RUN = "dry-run"
    SCHEDULED_COLLECTION = "scheduled"

    def __init__(
        self, collection_type=DRY_RUN, collector_module=None, logger=None, licensed=True
    ):
        self.licensed = licensed
        self.collector_module = collector_module
        self.collection_type = collection_type
        self.collections = {}
        self.packages = {}

        self.last_gathered_entries = None
        self.logger = logger or logging.getLogger(
            "insights-analytics-collector.collector"
        )
        self.log_level = (
            logging.ERROR
            if self.collection_type != self.SCHEDULED_COLLECTION
            else logging.DEBUG
        )

        self.tmp_dir = None
        self.gather_dir = None
        self.gather_since = None
        self.gather_until = None
        self.last_gather = None

    #
    # Class methods -----------------------------
    #
    @classmethod
    def registered_collectors(cls, module):
        """
        Returns all functions in 'module' defined with "@register" decorator
        """
        return {
            func.__insights_analytics_key__: {
                "name": func.__insights_analytics_key__,
                "version": func.__insights_analytics_version__,
                "description": func.__insights_analytics_description__ or "",
            }
            for name, func in inspect.getmembers(module)
            if inspect.isfunction(func) and hasattr(func, "__insights_analytics_key__")
        }

    #
    # Public methods ----------------------------
    #
    def config_present(self):
        """
        Checks if collector_module contains 'config' method (required)
        :return: bool
        """

        return self.collections.get("config") is not None

    @staticmethod
    @abstractmethod
    def db_connection():
        """
        DB connection for advisory lock. Can be
        - django.db.connection or
        - sqlalchemy.engine.base.Engine.raw_connection()
        - etc.
        """
        pass

    def gather(self, dest=None, subset=None, since=None, until=None):
        """Entry point for gathering

        :param dest: (default: /tmp/awx-analytics-*) - directory for temp files
        :param subset: (list) collector_module's function names if only subset is required (typically tests)
        :param since: (datetime) - low threshold of data changes (max. and default - 4 weeks ago)
        :param until: (datetime) - high threshold of data changes (defaults to now)
        :return: None or list of paths to tarballs (.tar.gz)
        """
        if not self.is_enabled():
            return None

        with self._pg_advisory_lock("gather_analytics_lock", wait=False) as acquired:
            if not acquired:
                self.logger.log(
                    self.log_level, "Not gathering analytics, another task holds lock"
                )
                return None

            self._gather_initialize(dest, subset, since, until)

            if not self._gather_config():
                return None

            self._gather_json_collections()

            self._gather_csv_collections()

            self._process_packages()

            self._gather_finalize()

            self._gather_cleanup()

            return self.all_tar_paths()

    def is_dry_run(self):
        return self.collection_type == self.DRY_RUN

    def is_enabled(self):
        """Checks for license and shipping data (like credentials)"""
        if not self._is_valid_license() and self.licensed:
            self.logger.log(
                self.log_level, "Invalid License provided, or No License Provided"
            )
            return False

        if self.is_shipping_enabled():
            return self._is_shipping_configured()

        return True

    def is_shipping_enabled(self):
        """Shipping is enabled in manual/scheduled mode"""
        return not self.is_dry_run()

    def last_gathered_entry_for(self, key):
        return self.last_gathered_entries.get(key)

    def all_tar_paths(self):
        tar_paths = []
        for _, packages in self.packages.items():
            new_paths = [
                package.tar_path for package in packages if package.tar_path is not None
            ]
            tar_paths += new_paths
        return tar_paths or []

    def delete_tarballs(self):
        for path in self.all_tar_paths():
            os.remove(path)

    #
    # Private methods ---------------------------
    #
    def _calculate_collection_interval(self, since, until):
        _now = now()

        # Make sure that the endpoints are not in the future.
        if until is not None and until > _now:
            until = _now
            self.logger.warning(
                f"End of the collection interval is in the future, setting to {_now}."
            )
        if since is not None and since > _now:
            since = _now
            self.logger.warning(
                f"Start of the collection interval is in the future, setting to {_now}."
            )

        # The value of `until` needs to be concrete, so resolve it.  If it wasn't passed in,
        # set it to `now`, but only if that isn't more than 4 weeks ahead of a passed-in
        # `since` parameter.
        if since is not None:
            if until is not None:
                if until > since + timedelta(weeks=4):
                    until = since + timedelta(weeks=4)
                    self.logger.warning(
                        f"End of the collection interval is greater than 4 weeks from start, setting end to {until}."
                    )
            else:  # until is None
                until = min(since + timedelta(weeks=4), _now)
        elif until is None:
            until = _now

        if since and since >= until:
            self.logger.warning(
                "Start of the collection interval is later than the end, ignoring request."
            )
            raise ValueError

        # The ultimate beginning of the interval needs to be compared to 4 weeks prior to
        # `until`, but we want to keep `since` empty if it wasn't passed in because we use that
        # case to know whether to use the bookkeeping settings variables to decide the start of
        # the interval.
        horizon = until - timedelta(weeks=4)
        if since is not None and since < horizon:
            since = horizon
            self.logger.warning(
                f"Start of the collection interval is more than 4 weeks prior to {until}, setting to {horizon}."
            )

        last_gather = self._last_gathering() or horizon
        if last_gather < horizon:
            last_gather = horizon
            self.logger.warning(
                f"Last analytics run was more than 4 weeks prior to {until}, using {horizon} instead."
            )

        self.gather_since = since
        self.gather_until = until
        self.last_gather = last_gather

    def _find_available_package(self, group, key, requested_size=None):
        """Checks if there is a Package available for collection.
        Package can't contain collection with the same key and has to have enough free space

        :param group: finds or creates package for group strategy if not None
        :param requested_size: returns existing package, if there is enough free size

        :return: Package
        """
        available_package = None

        for package in self.packages.get(group) or []:
            if (
                package.has_free_space(requested_size)
                and not package.is_key_used(key)
                and not package.processed
            ):
                available_package = package
                break

        if available_package is None:
            available_package = self._create_package()
            self.packages[group] = self.packages.get(group) or []
            self.packages[group].append(available_package)

        return available_package

    def _gather_initialize(self, tmp_root_dir, collectors_subset, since, until):
        self.logger.debug(f"Last analytics run was: {self._last_gathering()}")

        self._init_tmp_dir(tmp_root_dir)

        self.last_gathered_entries = self._load_last_gathered_entries()

        self._calculate_collection_interval(since, until)

        self._reset_collections_and_packages()

        self._create_collections(collectors_subset)

    def _gather_config(self):
        """Config is special collection, it's added to each Package
        TODO: add "always" flag to @register decorator
        """
        if not self.config_present():
            self.logger.log(self.log_level, "'config' collector data is missing")
            return False
        else:
            self.collections["config"].gather(self._package_class().max_data_size())
            return True

    def _gather_json_collections(self):
        """JSON collections are simpler, they're just gathered and added to the Package"""
        for collection in self.collections[Collection.COLLECTION_TYPE_JSON]:
            collection.gather(self._package_class().max_data_size())

            self._add_collection_to_package(collection)

    def _gather_csv_collections(self):
        """CSV collections can contain sub-collections (big db tables).
        In that case they are shipped immediately, because:
         1) the temp file needs to be deleted to ensure enough disk space
         2) Collections with slicing function can produce duplicate filename
        """
        for collection in self.collections[Collection.COLLECTION_TYPE_CSV]:
            collection.gather(self._package_class().max_data_size())

            if collection.is_empty() or not collection.gathering_successful:
                continue

            # If collection has sub_collections (it means it collected more files)
            # ship them in their own package
            if len(collection.sub_collections):
                for sub_collection in collection.sub_collections:
                    self._add_collection_to_package(sub_collection)
            else:
                self._add_collection_to_package(collection)

    def _add_collection_to_package(self, collection):
        """Adds collection to package and ships it if collection has slicing"""
        package = self._find_available_package(
            collection.shipping_group, collection.key, collection.data_size()
        )
        package.add_collection(collection)
        if collection.ship_immediately():
            self._process_package(package)

    @contextlib.contextmanager
    def _pg_advisory_lock(self, key, wait=False):
        """Postgres db lock"""
        connection = self.db_connection()

        if connection is None:
            yield True
        else:
            # Build 64-bit integer out of the resource id
            resource_key = int(hashlib.sha512(key.encode()).hexdigest(), 16) % 2**63

            cursor = connection.cursor()

            try:
                if wait:
                    cursor.execute("SELECT pg_advisory_lock(%s);", (resource_key,))
                else:
                    cursor.execute("SELECT pg_try_advisory_lock(%s);", (resource_key,))
                acquired = cursor.fetchall()[0][0]
                yield acquired
            finally:
                cursor.execute("SELECT pg_advisory_unlock(%s);", (resource_key,))
                cursor.close()

    def _process_packages(self):
        for group, packages in self.packages.items():
            for package in packages:
                self._process_package(package)

    def _process_package(self, package):
        """
        Processing of package can be called twice, skipping the 2nd call.
        If there is a custom slicing function,
        package has to be sent immediately after gathering data
        :see Collection.ship_immediately()

        :param package: Package
        """
        if not package.processed:
            package.make_tgz()
            if self.is_shipping_enabled():
                package.ship()
            package.delete_collected_files()
            package.processed = True

    def _gather_finalize(self):
        """Persisting timestamps (manual/schedule mode only)"""
        if self.is_shipping_enabled():
            self._update_last_gathered_entries()

            self._save_last_gather()

    def _gather_cleanup(self):
        """Deleting temp files"""
        shutil.rmtree(
            self.tmp_dir, ignore_errors=True
        )  # clean up individual artifact files
        if not self.is_dry_run():
            self.delete_tarballs()

    def _init_tmp_dir(self, tmp_root_dir=None):
        self.tmp_dir = pathlib.Path(
            tmp_root_dir or tempfile.mkdtemp(prefix="awx_analytics-")
        )
        self.gather_dir = self.tmp_dir.joinpath("stage")
        self.gather_dir.mkdir(mode=0o700)

    @abstractmethod
    def _is_shipping_configured(self):
        """Custom check for shipping availability should contain:
        1) Is Insights for Ansible Automation Platform enabled?
        2) Is URL and credentials present?
        :return: bool
        """
        pass

    @abstractmethod
    def _is_valid_license(self):
        """License check
        :return: bool
        """
        pass

    @abstractmethod
    def _last_gathering(self):
        """Returns timestamp of last successful gathering
        Complement to _save_last_gathering()
        """
        pass

    @abstractmethod
    def _load_last_gathered_entries(self):
        """Loads persisted timestamps named by keys from collector_module
        Complement to the _save_last_gathered_entries()
        :return dict
        """
        pass

    def _update_last_gathered_entries(self):
        last_gathered_updates = {"keys": {}, "locked": set()}

        for _, packages in self.packages.items():
            for package in packages:
                package.update_last_gathered_entries(last_gathered_updates)

        # Locked key means that gathering wasn't successful at least once.
        # Full sync timestamp can't be updated (if present)
        for unsuccessful_key in last_gathered_updates["locked"]:
            last_gathered_updates.pop(f"{unsuccessful_key}_full", None)

        self.last_gathered_entries.update(last_gathered_updates["keys"])

        self._save_last_gathered_entries(self.last_gathered_entries)

    @abstractmethod
    def _save_last_gathered_entries(self, last_gathered_entries):
        """Saves dictionary with timestamps to persistent storage
        Complement to the _load_last_gathered_entries()
        :param last_gathered_entries: dict
        """
        pass

    @abstractmethod
    def _save_last_gather(self):
        """Persists timestamp of last successful gathering
        Complement to _last_gathering()
        """
        pass

    def _create_collections(self, subset=None):
        """Creates Collections from decorated functions (by @register) from self.collector_module
        :param subset - array of function names which should be used.
                      - if None, all registered functions will be used

        collections have following structure:
        {
            'json': []
            'csv': []
            'config': <Collection>
        }
        """
        for name, fnc in inspect.getmembers(self.collector_module):
            if (
                inspect.isfunction(fnc)  # noqa
                and hasattr(fnc, "__insights_analytics_key__")  # noqa
                and hasattr(fnc, "__insights_analytics_type__")  # noqa
                and (not subset or name in subset)  # noqa
            ):
                # Create collection by type
                collection = self._create_collection(fnc)

                if collection.is_config:
                    # It's supposed there is only one registered config
                    self.collections[Collection.COLLECTION_TYPE_CONFIG] = collection
                else:
                    for since, until in collection.slices():
                        collection.since = since
                        collection.until = until
                        self.collections[collection.data_type].append(collection)
                        collection = self._create_collection(fnc)

    def _create_collection(self, fnc_collecting):
        data_type = fnc_collecting.__insights_analytics_type__
        collection = None
        if data_type == "json":
            collection = self._collection_json_class()(self, fnc_collecting)
        elif data_type == "csv":
            collection = self._collection_csv_class()(self, fnc_collecting)

        if collection is None:
            raise RuntimeError(f"Collection of type {data_type} not implemented")

        return collection

    def _create_package(self):
        package_class = self._package_class()
        return package_class(self)

    @staticmethod
    def _package_class():
        """Has to be redefined by your Package implementation"""
        return Package

    @staticmethod
    def _collection_json_class():
        """Can be redefined by your CollectionJSON implementation"""
        return CollectionJSON

    @staticmethod
    def _collection_csv_class():
        """Can be redefined by your CollectionCSV implementation"""
        return CollectionCSV

    @staticmethod
    def collection_data_status_class():
        return CollectionDataStatus

    @staticmethod
    def collection_manifest_class():
        return CollectionManifest

    def _reset_collections_and_packages(self):
        self.collections = {
            Collection.COLLECTION_TYPE_JSON: [],
            Collection.COLLECTION_TYPE_CSV: [],
            Collection.COLLECTION_TYPE_CONFIG: None,
        }
        self.packages = {}



from .collection_json import CollectionJSON
from .decorators import register


class CollectionManifest(CollectionJSON):
    def __init__(self, collector):
        super().__init__(collector, self.collecting)

        self.data = {}
        self.since = collector.gather_since
        self.until = collector.gather_until

    @register("manifest", "1.0", format="json", description="Manifest file")
    def collecting(self, **kwargs):
        """Collecting function is skipped"""
        return self.data

    def add_collection(self, collection):
        self.data[collection.filename] = collection.version
